////////////////////////////////////////////////////////////////////////////////
//3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
//      10        20        30        40        50        60        70        80
//
// gl-particles
//
// main.c - pull everything together
//
// Authors/Copyright 2009:
//    Mirco "MacSlow" Müller <macslow@bangang.de>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gtk/gtkgl.h>

#include "events.h"

#define FPS 1000/60

int 
main (int    argc,
      char** argv)
{
	GtkWidget*   window             = NULL;
	GdkGLConfig* gl_config          = NULL;
	guint        draw_handler_id    = 0;
	guint        pointer_handler_id = 0;
	gint         attrib_list[]      = {GDK_GL_USE_GL,       1,
					   GDK_GL_RGBA,         1,
					   GDK_GL_DOUBLEBUFFER, 1,
					   GDK_GL_RED_SIZE,     8,
					   GDK_GL_GREEN_SIZE,   8,
					   GDK_GL_BLUE_SIZE,    8,
					   GDK_GL_ALPHA_SIZE,   8,
					   GDK_GL_DEPTH_SIZE,  24,
					   GDK_GL_STENCIL_SIZE, 8,
					   GDK_GL_ATTRIB_LIST_NONE};

	gtk_init (&argc, &argv);
	gtk_gl_init (&argc, &argv);

	gl_config = gdk_gl_config_new (attrib_list);

	if (!gl_config)
	{
		g_print ("Could not create OpenGL-config\n!");
		exit (1);
	}

	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	if (!window)
	{
		g_print ("Could not create GTK+-window\n!");
		exit (2);
	}

	gtk_widget_set_gl_capability (window,
				      gl_config,
				      NULL,
				      FALSE,
				      GDK_GL_RGBA_TYPE);

	g_signal_connect_after (G_OBJECT (window),
				"realize",
				G_CALLBACK (on_window_realize),
				NULL);
	g_signal_connect (G_OBJECT (window),
			  "configure-event",
			  G_CALLBACK (on_window_configure),
			  NULL);
	g_signal_connect (G_OBJECT (window),
			  "expose-event",
			  G_CALLBACK (on_window_expose),
			  NULL);
	g_signal_connect (G_OBJECT (window),
			  "key-press-event",
			  G_CALLBACK (on_key_press),
			  NULL);
	g_signal_connect (G_OBJECT (window),
			  "button-press-event",
			  G_CALLBACK (on_button),
			  NULL);
	g_signal_connect (G_OBJECT (window),
			  "button-release-event",
			  G_CALLBACK (on_button),
			  NULL);

	gtk_widget_add_events (window,
			       GDK_POINTER_MOTION_MASK | 
			       GDK_BUTTON_PRESS_MASK |
			       GDK_BUTTON_RELEASE_MASK);
	gtk_window_set_title (GTK_WINDOW (window), "particles with OpenGL");
	gtk_window_set_icon_from_file (GTK_WINDOW (window),
				       "./opengl-logo.svg",
				       NULL);
	gtk_widget_set_size_request (window, 500, 500);
	//gtk_window_move (GTK_WINDOW (window), 220, 140);
	gtk_widget_show_all (window);

	g_signal_connect (G_OBJECT (window),
			  "delete-event",
			  G_CALLBACK (on_delete),
			  NULL);

	// force a refresh-rate of 60 Hz
	draw_handler_id = g_timeout_add (FPS,
					 (GSourceFunc) draw_handler,
					 window);

	// mouse-pointer monitoring
	pointer_handler_id = g_timeout_add (FPS,
					    (GSourceFunc) pointer_handler,
					    window);

	gtk_main ();

	g_source_remove (draw_handler_id);
	g_source_remove (pointer_handler_id);

	return 0;
}

