////////////////////////////////////////////////////////////////////////////////
//3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
//      10        20        30        40        50        60        70        80
//
// gl-particles
//
// opengl.c - all the OpenGL-related drawning code you'll find here
//
// Authors/Copyright 2009:
//    Mirco "MacSlow" Müller <macslow@bangang.de>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//
////////////////////////////////////////////////////////////////////////////////

#include <gtk/gtk.h>
#include <gtk/gtkgl.h>
#include <gdk/gdkx.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <math.h>

#include "opengl.h"
#include "particle.h"

#define Z_NEAR           0.1f
#define Z_FAR          100.0f
#define FOVY            35.0f
#define MAX_PARTICLES 5000

GLdouble  g_camera_rot[3] = {169.5f, 246.5f,  0.0f};  // camera's pitch, yaw, roll
GLdouble  g_camera_pos[3] = {  3.94f,  3.71f, 4.97f}; // camera-center/eye
GLdouble  g_camera_aim[3] = {  0.0f,   0.0f,  0.0f};  // camera's target-point
GLdouble  g_camera_up[3]  = {  0.0f,   1.0f,  0.0f};  // camera's up-vector
GLdouble  g_direction[3]  = {  0.0f,   0.0f,  1.0f};  // normalized view-vector
GLuint    g_coord_list_id = 0;                        // display-list id
GRand*    g_rand          = NULL;                     // rand. number-generator
Particle* g_particles[MAX_PARTICLES];                 // list of particles

void
_draw_coord_sys ()
{
	GLint i;
	GLint j;

	glPushMatrix ();
	glBegin (GL_LINES);

	glLineWidth (0.5f);

	// draw floor
	glColor4f (0.5f, 0.5f, 0.5f, 0.25f);
	for (i = -50; i < 50; i++)
	{
		for (j = -50; j < 50; j++)
		{
			if ((i >= -2 && i <= 1) &&
			    (j >= -2 && j <= 1))
				continue;

			glVertex3f (i * 0.5f,        0.0f, j * 0.5f);
			glVertex3f (i * 0.5f + 0.5f, 0.0f, j * 0.5f);
			glVertex3f (i * 0.5f,        0.0f, j * 0.5f);
			glVertex3f (i * 0.5f,        0.0f, j * 0.5f + 0.5f);
		}
	}

	glLineWidth (1.5f);

	// draw pos. x-axis
	glColor3f (1.0f, 0.0f, 0.0f);
	glVertex3f (0.0f, 0.0f, 0.0f);
	glVertex3f (1.0f, 0.0f, 0.0f);

	// draw neg. x-axis
	for (i = 0; i < 10; i++)
	{
		glVertex3f (-0.1f * i,         0.0f, 0.0f);
		glVertex3f (-0.1f * i - 0.01f, 0.0f, 0.0f);
	}
	
	// draw pos. y-axis
	glColor3f (0.0f, 1.0f, 0.0f);
	glVertex3f (0.0f, 0.0f, 0.0f);
	glVertex3f (0.0f, 1.0f, 0.0f);

	// draw neg. y-axis
	for (i = 0; i < 10; i++)
	{
		glVertex3f (0.0f, -0.1f * i,         0.0f);
		glVertex3f (0.0f, -0.1f * i - 0.01f, 0.0f);
	}

	// draw pos. z-axis
	glColor3f (0.0f, 0.0f, 1.0f);
	glVertex3f (0.0f, 0.0f, 0.0f);
	glVertex3f (0.0f, 0.0f, 1.0f);

	// draw neg. z-axis
	for (i = 0; i < 10; i++)
	{
		glVertex3f (0.0f, 0.0f, -0.1f * i);
		glVertex3f (0.0f, 0.0f, -0.1f * i - 0.01f);
	}

	glEnd ();
	glPopMatrix ();
}

void
_recalc_camera ()
{
	GLdouble length = 1.0f;

	// pitch
	g_camera_aim[Y] = g_camera_pos[Y] + 1.0f * cos (G_PI / 180.0f * g_camera_rot[X]);
	g_camera_aim[Z] = g_camera_pos[Z] + 1.0f * sin (G_PI / 180.0f * g_camera_rot[X]);

	// yaw
	g_camera_aim[X] = g_camera_pos[X] + 1.0f * cos (G_PI / 180.0f * g_camera_rot[Y]);
	g_camera_aim[Z] = g_camera_pos[Z] + 1.0f * sin (G_PI / 180.0f * g_camera_rot[Y]);

	// update view-vector
	g_direction[X] = g_camera_pos[X] - g_camera_aim[X];
	g_direction[Y] = g_camera_pos[Y] - g_camera_aim[Y];
	g_direction[Z] = g_camera_pos[Z] - g_camera_aim[Z];

	// normalize view-vector
	length = sqrt (g_direction[X] * g_direction[X] +
		       g_direction[Y] * g_direction[Y] +
		       g_direction[Z] * g_direction[Z]);
	if (length != 0.0f)
	{
		g_direction[X] /= length;
		g_direction[Y] /= length;
		g_direction[Z] /= length;
	}

	/*g_print ("pos: %3.2f/%3.2f/%3.2f\naim: %3.2f/%3.2f/%3.2f\nrot: %3.2f/%3.2f/%3.2f\ndir: %3.2f/%3.2f/%3.2f\n\n",
		 g_camera_pos[X],
		 g_camera_pos[Y],
		 g_camera_pos[Z],
		 g_camera_aim[X],
		 g_camera_aim[Y],
		 g_camera_aim[Z],
		 g_camera_rot[X],
		 g_camera_rot[Y],
		 g_camera_rot[Z],
		 g_direction[X],
		 g_direction[Y],
		 g_direction[Z]);*/

	//g_camera_up[Y]  = g_camera_pos[Y] + cos (G_PI / 180.0f * (g_camera_rot[X] + 90.0f));
	//g_camera_up[Z]  = g_camera_pos[Z] + sin (G_PI / 180.0f * (g_camera_rot[X] + 90.0f));
}

void
_randomize_particle (Particle* particle)
{
	if (!particle)
		return;

	if (!g_rand)
		return;

	// wall-emitter
	particle->position[X]  = g_rand_double_range (g_rand, -0.5f, 0.5f);
	particle->position[Y]  = g_rand_double_range (g_rand,  1.0f, 1.3f);
	particle->position[Z]  = -1.0f;
	particle->direction[X] = g_rand_double_range (g_rand, -0.2f, 0.2f);
	particle->direction[Y] = g_rand_double_range (g_rand, -0.2f, 0.2f);
	particle->direction[Z] = g_rand_double_range (g_rand, 0.01f, 1.0f);
	particle->speed        = g_rand_double_range (g_rand, 0.05f, 0.1f);
	particle->age          = 0.0f;
	particle->lifetime     = g_rand_double_range (g_rand,
						      1.0f,
						      0.1f * 60.0f * 2.0f);

	// singularity-emitter
	/*particle->position[X]  = 0.0f;
	particle->position[Y]  = 0.0f;
	particle->position[Z]  = 0.0f;
	particle->direction[X] = g_rand_double_range (g_rand, -0.5f, 0.5f);
	particle->direction[Y] = g_rand_double_range (g_rand, -0.5f, 0.5f);
	particle->direction[Z] = g_rand_double_range (g_rand, -0.5f, 0.5f);
	particle->speed        = g_rand_double_range (g_rand, 0.03f, 0.06f);
	particle->age          = 0.0f;
	particle->lifetime     = g_rand_double_range (g_rand,
						      1.0f,
						      0.1f * 60.0f * 1.0f);*/

}

void
_apply_gravity (Particle* particle)
{
	if (!particle)
		return;

	particle->direction[Y] += -0.1f;
	particle->speed        += 0.001f;
}

void
_create_particles ()
{
	int    i;
	double position[3]  = {0.0f, 0.0f, 0.0f};
	double direction[3] = {0.0f, 0.0f, 0.0f};
	double speed        = 1.0f;
	double lifetime     = 1.0f;

	if (!g_rand)
		return;

	for (i = 0; i < MAX_PARTICLES; i++)
	{
		g_particles[i] = particle_create (position,
						  direction,
						  speed,
						  lifetime);
		_randomize_particle (g_particles[i]);
	}
}

void
_update_particles ()
{
	int i;

	for (i = 0; i < MAX_PARTICLES; i++)
	{
		if (g_particles[i])
		{
			if (g_particles[i]->age < g_particles[i]->lifetime)
			{
				_apply_gravity (g_particles[i]);
				particle_update (g_particles[i]);
			}
			else
				_randomize_particle (g_particles[i]);
		}			
	}
}

void
_draw_particles ()
{
	int i;

	glBegin (GL_POINTS);
	for (i = 0; i < MAX_PARTICLES; i++)
	{
		if (g_particles[i])
		{
			glColor4f (0.0f,
				   0.0f,
				   0.0f,
				   1.0f - g_particles[i]->age /
				   g_particles[i]->lifetime);
			glVertex3f (g_particles[i]->position[X],
				    g_particles[i]->position[Y],
				    g_particles[i]->position[Z]);
		}
	}
	glEnd ();
}

void
_destroy_particles ()
{
	int i;

	for (i = 0; i < MAX_PARTICLES; i++)
		particle_destroy (g_particles[i]);
}

void
opengl_set_blend_mode (int mode)
{
	switch (mode)
	{
		// normal
		case 1:
			glBlendFunc (GL_SRC_ALPHA,
				     GL_ONE_MINUS_SRC_ALPHA);
			glTexEnvi (GL_TEXTURE_ENV,
				   GL_TEXTURE_ENV_MODE,
				   GL_DECAL);
			glTexEnvi (GL_TEXTURE_ENV,
				   GL_TEXTURE_ENV_COLOR,
				   GL_DECAL);
		break;

		// modulate
		case 2:
			glBlendFunc (GL_DST_COLOR,
				     GL_ZERO);
			glTexEnvi (GL_TEXTURE_ENV,
				   GL_TEXTURE_ENV_MODE,
				   GL_MODULATE);
			glTexEnvi (GL_TEXTURE_ENV,
				   GL_TEXTURE_ENV_COLOR,
				   GL_MODULATE);
		break;

		default :
		break;
	}
}

GLenum
opengl_what_texture_target (GLuint iWidth,
			    GLuint iHeight)
{
	GLenum   iTarget      = 0;
	gboolean bWidthIsPOT  = FALSE;
	gboolean bHeightIsPOT = FALSE;
	gint     i            = 1;

	// test width for power-of-two
	for (i = 1; i <= 32; i++)
		if ((unsigned long) iWidth - (1L << i) == 0)
		{
			bWidthIsPOT = TRUE;
			break;
		}

	// test height for power-of-two
	for (i = 1; i <= 32; i++)
		if ((unsigned long) iHeight - (1L << i) == 0)
		{
			bHeightIsPOT = TRUE;
			break;
		}

	// final conclusion
	if (bWidthIsPOT && bHeightIsPOT)
		iTarget = GL_TEXTURE_2D;
	else
		iTarget = GL_TEXTURE_RECTANGLE_ARB;

	return iTarget;	
}

gboolean
opengl_texture_proxy_check (GLenum iTarget,
			    gint   iChannels,
			    gint   iFormat,
			    gint   iWidth,
			    gint   iHeight)
{
	GLenum iProxyTarget = 0;
	GLint  iProxyWidth  = 0;
	GLint  iProxyHeight = 0;

	// select appropriate proxy-target
	switch (iTarget)
	{
		case GL_TEXTURE_1D :
			iProxyTarget = GL_PROXY_TEXTURE_1D;
			glTexImage1D (iProxyTarget,
				      0,
				      iChannels,
				      iWidth,
				      0,
				      iFormat,
				      GL_UNSIGNED_BYTE,
				      NULL);
		break;

		case GL_TEXTURE_2D :
			iProxyTarget = GL_PROXY_TEXTURE_2D;
			glTexImage2D (iProxyTarget,
				      0,
				      iChannels,
				      iWidth,
				      iHeight,
				      0,
				      iFormat,
				      GL_UNSIGNED_BYTE,
				      NULL);
		break;

		case GL_TEXTURE_RECTANGLE_ARB :
			iProxyTarget = GL_PROXY_TEXTURE_RECTANGLE_ARB;
			glTexImage2D (iProxyTarget,
				      0,
				      iChannels,
				      iWidth,
				      iHeight,
				      0,
				      iFormat,
				      GL_UNSIGNED_BYTE,
				      NULL);
		break;
	}

	/* query the results from the proxy-texture-target */
	glGetTexLevelParameteriv (iProxyTarget,
				  0,
				  GL_TEXTURE_WIDTH,
				  &iProxyWidth);
	glGetTexLevelParameteriv (iProxyTarget,
				  0,
				  GL_TEXTURE_HEIGHT,
				  &iProxyHeight);

	/* do the final check if texture would have fitted within the limits */
	if (iWidth == iProxyWidth &&
	    iHeight == iProxyHeight)
		return TRUE;
	else
		return FALSE;
}

Texture*
opengl_upload_texture (gchar* pcFilename)
{
	GLuint	   uiTextureId	   = 0;
	GdkPixbuf* pSrcPixbuf	   = NULL;
	GError*	   pError	   = NULL;
	gint	   iBitsPerChannel = 0;
	gint	   iWidth	   = 0;
	gint	   iHeight	   = 0;
	guchar*	   pucPixelBuffer  = NULL;
	gint	   iChannels	   = 0;
	Texture*   pTexture	   = NULL;

	/* sanity check */
	g_assert (pcFilename != NULL);

	/* create a texture-object */
	pTexture = (Texture*) g_malloc0 (sizeof (Texture));
	if (!pTexture)
		return NULL;

	/* create a GL texture-handle */
	glGenTextures (1, &uiTextureId);
	if (uiTextureId == 0)
	{
		g_free ((gpointer) pTexture);
		return NULL;
	}

	/* load the image from disk */
	pSrcPixbuf = gdk_pixbuf_new_from_file (pcFilename, &pError);
	if (pSrcPixbuf == NULL)
	{
		g_free ((gpointer) pTexture);
		glDeleteTextures (1, &uiTextureId);
		return NULL;
	}

	/* only allow 8 bit per color-component */
	iBitsPerChannel = gdk_pixbuf_get_bits_per_sample (pSrcPixbuf);
	if (iBitsPerChannel != 8)
	{
		g_object_unref (pSrcPixbuf);
		g_free ((gpointer) pTexture);
		glDeleteTextures (1, &uiTextureId);
		return NULL;
	}

	/* determine the number of components */
	iChannels = gdk_pixbuf_get_n_channels (pSrcPixbuf);

	/* get hold of dimensions and pixel-data */
	iWidth = gdk_pixbuf_get_width (pSrcPixbuf);
	iHeight = gdk_pixbuf_get_height (pSrcPixbuf);
	pucPixelBuffer = gdk_pixbuf_get_pixels (pSrcPixbuf);

	pTexture->uiTextureId	= uiTextureId;

	pTexture->fWidth  = (GLfloat) iWidth;
	pTexture->fHeight = (GLfloat) iHeight;

	pTexture->iTarget	= opengl_what_texture_target (pTexture->fWidth,
							      pTexture->fHeight);
	pTexture->iBufferSize	= 0;
	pTexture->pucBufferData	= NULL;

	pTexture->fMinS = 0.0f;
	pTexture->fMinT = 0.0f;
	pTexture->fMaxS = (pTexture->iTarget == GL_TEXTURE_2D) ? 1.0f : pTexture->fWidth;
	pTexture->fMaxT = (pTexture->iTarget == GL_TEXTURE_2D) ? 1.0f : pTexture->fHeight;

	/* see if the texture would fit */
	if (!opengl_texture_proxy_check (pTexture->iTarget,
					 iChannels,
					 (iChannels == 4) ? GL_RGBA : GL_RGB,
					 (GLint) pTexture->fWidth,
					 (GLint) pTexture->fHeight))
	{
		g_object_unref (pSrcPixbuf);
		g_free ((gpointer) pTexture);
		glDeleteTextures (1, &pTexture->uiTextureId);
		return NULL;
	}

	/* finally create the GL texture-object */
	glBindTexture (pTexture->iTarget, pTexture->uiTextureId);
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MIN_FILTER,
			 GL_LINEAR);
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MAG_FILTER,
			 GL_LINEAR);
	glTexImage2D (pTexture->iTarget,
		      0,
		      iChannels,
		      (GLsizei) pTexture->fWidth,
		      (GLsizei) pTexture->fHeight,
		      0,
		      (iChannels == 4) ? GL_RGBA : GL_RGB,
		      GL_UNSIGNED_BYTE,
		      pucPixelBuffer);

	g_object_unref (pSrcPixbuf);

	return pTexture;
}

void
opengl_delete_texture (Texture* texture)
{
	if (!texture);
		return;

	glDeleteTextures (1, &texture->uiTextureId);

	if (texture->pucBufferData)
		g_free ((gpointer) texture->pucBufferData);

	g_free ((gpointer) texture);
}

void
opengl_print_error (gchar* func_name)
{
	GLenum error;

	error = glGetError ();

	switch (error)
	{
		case GL_NO_ERROR :
			g_print ("%s() caused GL_NO_ERROR\n", func_name);
		break;

		case GL_INVALID_ENUM :
			g_print ("%s() caused GL_INVALID_ENUM\n", func_name);
		break;

		case GL_INVALID_VALUE :
			g_print ("%s() GL_INVALID_VALUE\n", func_name);
		break;

		case GL_INVALID_OPERATION :
			g_print ("%s() GL_INVALID_OPERATION\n", func_name);
		break;

		case GL_STACK_OVERFLOW :
			g_print ("%s() GL_STACK_OVERFLOW\n", func_name);
		break;

		case GL_STACK_UNDERFLOW :
			g_print ("%s() GL_STACK_UNDERFLOW\n", func_name);
		break;

		case GL_OUT_OF_MEMORY :
			g_print ("%s() GL_OUT_OF_MEMORY\n", func_name);
		break;

		case GL_TABLE_TOO_LARGE :
			g_print ("%s() GL_TABLE_TOO_LARGE\n", func_name);
		break;
	}
}

void
opengl_change_projection (gboolean parallel,
			  GLsizei  width,
			  GLsizei  height)
{
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	if (parallel)
		glOrtho ((GLdouble) -width / 2, (GLdouble) width / 2,
			 (GLdouble) -height / 2, (GLdouble) height / 2,
			 Z_NEAR, Z_FAR);
	else
		gluPerspective (2.0f * FOVY,
				(GLdouble) width / (GLdouble) height,
				Z_NEAR,
				Z_FAR);

	glMatrixMode (GL_MODELVIEW);
}

void
opengl_init (GLint width,
	     GLint height)
{
	GLfloat diffuse_light[4] = {0.76f, 0.75f, 0.65f, 1.0f};

	// set some GL-states
	glClearColor (1.0f, 0.5f, 0.25f, 1.0f);
	glColor4f (0.25f, 0.5f, 1.0f, 1.0f);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_RECTANGLE_ARB);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, 1);
	glEnable (GL_DEPTH_TEST);

	// make sure we get nice thick point-"sprites"
	glPointSize (2.5f);
	glEnable (GL_POINT_SMOOTH);
	glHint (GL_POINT_SMOOTH_HINT, GL_NICEST);

	// get nicely drawn lines
	glEnable (GL_LINE_SMOOTH);
	glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);

	// lights and so on
	glDisable (GL_LIGHTING);
	glDisable (GL_LIGHT0);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, diffuse_light);

	// setup "camera"
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	// store the coordinate-system object in a display-list if possible
	g_coord_list_id = 0;
	g_coord_list_id = glGenLists (1);
	if (g_coord_list_id)
	{
		glNewList (g_coord_list_id, GL_COMPILE);
		_draw_coord_sys ();
		glEndList ();
	}

	// initialize particle list
	g_rand = g_rand_new_with_seed (g_random_int ());
	_create_particles ();

	// position "camera"
	_recalc_camera ();
	gluLookAt (g_camera_pos[X], g_camera_pos[Y], g_camera_pos[Z],
		   g_camera_aim[X], g_camera_aim[Y], g_camera_aim[Z],
		   g_camera_up[X],  g_camera_up[Y],  g_camera_up[Z]);
}

void
opengl_resize (GLsizei width,
               GLsizei height)
{
	glViewport (0, 0, width, height);
	opengl_change_projection (FALSE, width, height);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void
opengl_draw (GtkWidget* window,
	     GLsizei    width,
	     GLsizei    height)
{
	GtkStyle*      style = NULL;
	GLfloat        bg_color[3];
	static GLfloat angle = 0.0f;

	// acquire theme's background color
	style = gtk_widget_get_style (window);
	bg_color[R] = (GLfloat) style->bg[GTK_STATE_NORMAL].red;
	bg_color[G] = (GLfloat) style->bg[GTK_STATE_NORMAL].green;
	bg_color[B] = (GLfloat) style->bg[GTK_STATE_NORMAL].blue;
	bg_color[R] /= (GLfloat) 0xFFFF;
	bg_color[G] /= (GLfloat) 0xFFFF;
	bg_color[B] /= (GLfloat) 0xFFFF;

	// clear background
	glClearColor (bg_color[R], bg_color[G], bg_color[B], 1.0f);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix ();
	glLoadIdentity ();
	gluLookAt (g_camera_pos[X], g_camera_pos[Y], g_camera_pos[Z],
		   g_camera_aim[X], g_camera_aim[Y], g_camera_aim[Z],
		   g_camera_up[X],  g_camera_up[Y],  g_camera_up[Z]);

	// draw coordinate-system
	if (g_coord_list_id)
		glCallList (g_coord_list_id);
	else
		_draw_coord_sys ();

	// draw particle system
	_update_particles ();
	_draw_particles ();

	// draw four points of a quad
	/*glPushMatrix ();
	glRotatef (angle, 0.0f, 0.0f, 1.0f);
	glBegin (GL_POINTS);
	glColor3f (0.0f, 0.0f, 0.0f);
	glVertex3f (-1.0f, -.2f, -1.0f);
	glVertex3f ( 1.0f, -.2f, -1.0f);
	glVertex3f ( 1.0f,  .2f, -1.0f);
	glVertex3f (-1.0f,  .2f, -1.0f);
	glEnd ();
	glPopMatrix ();*/

	glPopMatrix ();

	angle += 1.0f;
}

void
opengl_cleanup ()
{
	_destroy_particles ();
	g_rand_free (g_rand);
}

void
opengl_camera_reset ()
{
	g_camera_rot[X] = 169.5f;
	g_camera_rot[Y] = 246.5f;
	g_camera_rot[Z] =   0.0f;
	g_camera_pos[X] =   3.94f;
	g_camera_pos[Y] =   3.71f;
	g_camera_pos[Z] =   4.97f;
	g_camera_up[X]  =   0.0f;
	g_camera_up[Y]  =   1.0f;
	g_camera_up[Z]  =   0.0f;
	_recalc_camera ();
}

void
opengl_camera_move_forward ()
{
	g_camera_pos[X] -= 0.1f * g_direction[X];
	g_camera_pos[Y] -= 0.1f * g_direction[Y];
	g_camera_pos[Z] -= 0.1f * g_direction[Z];
	_recalc_camera ();
}

void
opengl_camera_move_backward ()
{
	g_camera_pos[X] += 0.1f * g_direction[X];
	g_camera_pos[Y] += 0.1f * g_direction[Y];
	g_camera_pos[Z] += 0.1f * g_direction[Z];
	_recalc_camera ();
}

void
opengl_camera_move_right ()
{
	g_camera_pos[X] += 0.1f * g_direction[Z];
	g_camera_pos[Z] += 0.1f * -g_direction[X];
	_recalc_camera ();
}

void
opengl_camera_move_left ()
{
	g_camera_pos[X] += 0.1f * -g_direction[Z];
	g_camera_pos[Z] += 0.1f * g_direction[X];
	_recalc_camera ();
}

void
opengl_camera_move_up ()
{
	g_camera_pos[Y] += 0.1f;
	_recalc_camera ();
}

void
opengl_camera_move_down ()
{
	g_camera_pos[Y] -= 0.1f;
	_recalc_camera ();
}

void
opengl_camera_pitch (GLdouble delta)
{
	g_camera_rot[X] -= delta;
	_recalc_camera ();
}

void
opengl_camera_yaw (GLdouble delta)
{
	g_camera_rot[Y] -= delta;
	_recalc_camera ();
}

