////////////////////////////////////////////////////////////////////////////////
//3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
//      10        20        30        40        50        60        70        80
//
// gl-particles
//
// opengl.h - header for all the OpenGL-related drawning code
//
// Authors/Copyright 2009:
//    Mirco "MacSlow" Müller <macslow@bangang.de>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//
////////////////////////////////////////////////////////////////////////////////

#ifndef _OPENGL_H
#define _OPENGL_H

#include <X11/Xlib.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include <gtk/gtk.h>

typedef enum
{
	R = 0,
	G,
	B,
	A
} Channel;

typedef enum
{
	X = 0,
	Y,
	Z
} Position;

typedef enum
{
	WIDTH = 0,
	HEIGHT
} Size;

typedef struct _Texture
{
	GLenum   iTarget;
	GLuint	 uiTextureId;
	GLuint	 uiUnit;
	GLfloat  fWidth;
	GLfloat  fHeight;
	GLfloat	 fMinS;
	GLfloat	 fMinT;
	GLfloat	 fMaxS;
	GLfloat	 fMaxT;
	GLuint   iBufferSize;
	GLubyte* pucBufferData; /* free as soon as no longer needed */
} Texture;

gboolean
opengl_enable_tfp (void);

void
opengl_set_blend_mode (int iMode);

GLenum
opengl_what_texture_target (GLuint iWidth,
			    GLuint iHeight);

gboolean
opengl_texture_proxy_check (GLenum iTarget,
			    gint   iChannels,
			    gint   iFormat,
			    gint   iWidth,
			    gint   iHeight);

Texture*
opengl_upload_texture (gchar* pcFilename);

void
opengl_delete_texture (Texture* pTexture);

void
opengl_print_error (gchar* pcFuncName);

void
opengl_change_projection (gboolean parallel,
                          GLsizei  width,
                          GLsizei  height);

void
opengl_init (GLint width,
	     GLint height);

void
opengl_resize (GLsizei width,
               GLsizei height);

void
opengl_draw (GtkWidget* window,
	     GLsizei    width,
             GLsizei    height);

void
opengl_cleanup (void);

void
opengl_camera_reset ();

void
opengl_camera_move_forward ();

void
opengl_camera_move_backward ();

void
opengl_camera_move_right ();

void
opengl_camera_move_left ();

void
opengl_camera_move_up ();

void
opengl_camera_move_down ();

void
opengl_camera_pitch (GLdouble delta);

void
opengl_camera_yaw (GLdouble delta);

#endif // _OPENGL_H

