////////////////////////////////////////////////////////////////////////////////
//3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
//      10        20        30        40        50        60        70        80
//
// gl-particles
//
// particle.c - the beef for the handling particle-system
//
// Authors/Copyright 2009:
//    Mirco "MacSlow" Müller <macslow@bangang.de>
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranties of
// MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.
//
////////////////////////////////////////////////////////////////////////////////

#include "opengl.h"
#include "particle.h"

Particle*
particle_create (gdouble* position,
		 gdouble* direction,
		 gdouble  speed,
		 gdouble  lifetime)
{
	Particle* particle = NULL;

	// sanity checks
	if (!position && !direction && lifetime <= 0.0f)
		return NULL;

	// allocate new particle
	particle = g_new0 (Particle, 1);
	if (!particle)
		return NULL;

	// initialize it
	particle->position[X]  = position[X];
	particle->position[Y]  = position[Y];
	particle->position[Z]  = position[Z];
	particle->direction[X] = direction[X];
	particle->direction[Y] = direction[Y];
	particle->direction[Z] = direction[Z];
	particle->speed        = speed;
	particle->age          = 0.0f;
	particle->lifetime     = lifetime;

	return particle;
}

gboolean
particle_update (Particle* particle)
{
	// sanity-check
	if (!particle)
		return FALSE;

	// is it time for particle to bite the dust?
	if (particle->age == particle->lifetime)
		particle_destroy (particle);

	// update particles parameters
	particle->position[X] += particle->speed * particle->direction[X];
	particle->position[Y] += particle->speed * particle->direction[Y];
	particle->position[Z] += particle->speed * particle->direction[Z];	
	particle->age         += 0.1f;

	return TRUE;
}

gboolean
particle_destroy (Particle* particle)
{
	// sanity check
	if (!particle)
		return FALSE;

	g_free (particle);
	particle = NULL;

	return TRUE;
}

